#!/usr/bin/env python
"""

    GTK Kinetic scroller
    (C) Copyright 2007 Anthony Maro
    http://tony.maro.net/
    
    Based significantly on GPL pygame code from Kagu:
    https://www.guardiani.us/projects/kagu
    Copyright (c) 2007 Jesse Guardiani <jesse@guardiani.us>
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.
   
   
   Version 1.1.0
   
   Changes:
   
   Version 1.1.0
       Enabled GTK double buffering
       Changed drawing of sprites to cache in a pixmap instead of a pixbuf 
         to improve performance.  Pixmaps are preconfigured for the screen
         depth and format, and render significantly faster
       Pre-draw the font onto the sprites when cachine to a pixmap for speed
       Tweaked friction and speed settings
       Removed most of the wedging of connect events since they weren't needed
       
   Version 1.0.0
       Initial Release
   

"""


import gtk, pango, pygtk, time, threading, gobject

class Rect(object):
    def __init__(self, xywh):
        """
        xywh must be a 2 or 4 tuple or a rect instance.
        """
        if isinstance(xywh, Rect):
            self.left = xywh.left
            self.top = xywh.top
            self.width = xywh.width
            self.height = xywh.height
        else:
            if len(xywh) == 4:
                self.left, self.top, self.width, self.height = (i for i in xywh)
            elif len(xywh) == 2:
                self.left, self.top, self.width, self.height = (0.0,0.0) + tuple(i for i in xywh)
                
    def set_right(self, s):
        self.left = s - self.width
        
    def get_right(self):
        return self.left + self.width    
    
    right = property(get_right, set_right)
    
    def set_bottom(self, s):
        self.top = s - self.height
        
    def get_bottom(self):
        return self.top + self.height
    
    bottom = property(get_bottom, set_bottom)
        
    def collidepoint(self, xy):
        """
        Test if a point intersects with this rect.
        """
        x,y = xy
        return x >= self.left and x <= (self.left + self.width) and y >= self.top and y <= (self.top + self.height)

    def intersects(self, other):
        """
        Test if a rect intersects with this rect.
        """
        if self.left > other.left+other.width: return False
        if self.top > other.top + other.height: return False
        if self.left + self.width < other.left: return False
        if self.top + self.height < other.top: return False
        return True 
                

class BaseWidget(object):
  def __init__(self):    
    self.image = None
    self.rect  = None

  def handle_event(self,event):
    ''' override me '''

  def is_idle(self):
    ''' override me '''
    return 1

  def update(self):
    ''' override me '''

  def draw(self,widget):
    return

  def close(self):
    ''' override me '''
    
def RGBToHTMLColor(rgb_tuple):
    hexcolor = '#%02x%02x%02x' % rgb_tuple
    return hexcolor    

class Scrollitem(BaseWidget):
    def __init__(self, caption, image, fontsize=20, fontcolor=(0,0,0), selected=False):
        BaseWidget.__init__(self)
        self.selected = selected
        self.rect = Rect((0,0,image.get_width(),image.get_height()))
        self.caption = caption
        self.fontsize = fontsize
        self.fontcolor = RGBToHTMLColor(fontcolor)
        self.focus = False
        
        self.image = image # A pixmap object
        self.cache = None
        
                
    def draw(self, widget):
        buffer = widget.window
        
        if self.cache == None:
            # Convert to a pixmap for faster drawing! 
            self.cache = gtk.gdk.Pixmap(buffer, self.rect.width, self.rect.height, -1)         
            self.cache.draw_pixbuf(None, self.image, 0,0,0,0,-1,-1)
            # Pre-render the caption font
            self.drawtext(widget)
        
        buffer.draw_drawable(widget.get_style().fg_gc[gtk.STATE_NORMAL], self.cache, 0,0,self.rect.left, self.rect.top,self.rect.width,self.rect.height)
        
                 

    def drawtext(self, widget):
        gc = widget.window.new_gc()
        if self.caption <> '':               
            context = widget.create_pango_context()
            mytext = pango.parse_markup('<span foreground="' + self.fontcolor + '">' + self.caption + '</span>', u'\x00')
            layout = widget.create_pango_layout(mytext[1])
            layout.set_attributes(mytext[0])
            desc = pango.FontDescription('Times ' + str(self.fontsize))
            layout.set_font_description(desc)
            lx, ly = layout.get_pixel_size()
            self.cache.draw_layout(gc, 
                                      (self.rect.width / 2) - (lx / 2), 
                                      (self.rect.height / 2) - (ly / 2), 
                                      layout)   
    
    def set_focus(self,focus):
        if focus != self.focus:
            if not focus:
                self.offset = 0
        self.focus = focus

class ScrollWidget(gtk.DrawingArea):
  ''' This widget wraps sprites with equal heights and 
  presents the user with an interface for scrolling and
  selecting these sprites'''
  def __init__(self):
    gtk.DrawingArea.__init__(self)
    #BaseWidget.__init__(self) # call Group initializer
    #self.rect       = Rect(rect) # this is where we draw our widget within the surface
    self.rect = None
    self.y_spacer   = 5 # height in pixels of whitespace between the top and bottom of sprites
    
    self.wrap       = True
    
    self.scrollable     = None
    self.sprites        = []
    self.sorted_sprites = []
    self.selected_index = None
    self.need_sort      = False
    self.selected_cb = None
    self.last_state     = "idle" # see self._change_state() for possibilities
    self.cur_state      = "idle"
    self.tap_event_threshold = 12
    self.tap_distance_threshold = 12
    self.visible_sprite_indexes = [] # optimization for is_mostly_visible()
    self.mousex = -1    
    self.mousey = -1
    self.smousex = -1
    self.smousey = -1   
    self.dirty = True   
    self.image = None # Stores my blit buffer
    self.colorgc = None
    self.backgroundcolor = '#FFFFFF'
    self.lastupdate = time.clock()
    
    self.clear()
    
    # Connect all my own personal events
    self.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.POINTER_MOTION_MASK | \
                    gtk.gdk.POINTER_MOTION_HINT_MASK | gtk.gdk.STRUCTURE_MASK | gtk.gdk.EXPOSURE_MASK)

    gtk.DrawingArea.connect(self, "button_press_event", self.clickbutton)
    gtk.DrawingArea.connect(self, "button_release_event", self.upbutton)
    gtk.DrawingArea.connect(self, "motion_notify_event", self.mousemove)
    gtk.DrawingArea.connect(self, "destroy",self.unrealized)
    gtk.DrawingArea.connect(self, "expose_event", self.expose)
    gtk.DrawingArea.connect(self, "realize", self.realized)
    gtk.DrawingArea.connect(self, "size_allocate", self.resized)
    
    self.timeoutobj = -1
    
      
    
  def resized(self, widget, event=None, data=None):
      # Force recreation of buffer and size rect
      self.rect = None
      self.image = None
      self.update()
      self.expose() 
      # Call user connect if necessary
      
  def selecteditem(self):
      myitem = self.sorted_sprites[self.selected_index]
      return myitem
  
  def connect(self, eventtype, callback):
      # Wedge myself between...
      if eventtype == 'scrollitem_click_event':          
          self.selected_cb = callback
          return
          # Pass through cause we don't do that...
      gtk.DrawingArea.connect(self, eventtype, callback)

  def unrealized(self, widget=None, data=None):
      ''' Call when we are unrealized to stop the drawing loop '''
      print "Unrealized"
      if self.timeoutobj <> -1:
          gobject.source_remove(self.timeoutobj)
          self.timeoutobj = -1
      self.sprites = []
      self.sorted_sprites = []
      self.visible_sprite_indexes = []
          

  def updateme(self):
      ''' Called once every .075 seconds to handle drawing '''
      while gtk.events_pending():
          # Process all other events first...
          gtk.main_iteration(block=True)   
             
      if not self.is_idle():
          if len(self.sprites) == 0: return True
          self.update()
          self.expose()
          
      return True # continue using the timeout callback
      
  def _makerect(self):
      ''' Create the widget rectangle used for calculations '''
      w,h = self.size_request()
      self.rect = Rect((0,0,w,h))

  def clear(self):
      self.offset     = 0
      self.distance   = 0
      self.mousedown  = 0
      self.speed      = 0
      self.num_events = 0
      self.friction   = 3
      self.friction_c = 0
      self.viewport_index = 0
      self._makerect()
      self.image = None
      self.scrollable     = None
      self.sprites        = []
      self.sorted_sprites = []
      self.selected_index = None
      self.need_sort      = False
      self.selected_cb = None
      self.last_state     = "idle" # see self._change_state() for possibilities
      self.cur_state      = "idle"
      self.tap_event_threshold = 12
      self.tap_distance_threshold = 12
      self.visible_sprite_indexes = [] # optimization for is_mostly_visible()
      self.mousex = -1    
      self.mousey = -1
      self.smousex = -1
      self.smousey = -1   
      self.dirty = True   
      
  def realized(self, widget=None, event=None, data=None):
      ''' Triggered when the widget is finally ready to draw '''
      self.update()
      self.expose()      
      self.timeoutobj = gobject.timeout_add(10,self.updateme) 
      
    
  def expose(self, widget=None, event=None, data=None):
    
    if self.get_style().white_gc == None: return # See if we are usable...
    self.window.begin_paint_rect((self.rect.top,self.rect.left,self.rect.width,self.rect.height))
    ''' Main drawing routine '''
    if self.rect == None:
        self._makerect()    

    if self.image == None:  # Create the buffer pixmap if necessary
        try:            
            self.image =  gtk.gdk.Pixmap(self.window, self.rect.width, self.rect.height, depth=-1)
            self.colorgc = None
            self._makerect()
        except:
            self.image = None # Couldn't do that yet - must not be realized
            return
    
    # Set the colorgc if not already set    
    if self.colorgc == None:
        cmap = self.image.get_colormap()
        color = cmap.alloc_color(self.backgroundcolor)
        self.colorgc = self.image.new_gc()
        self.colorgc.set_foreground(color)
        self.colorgc.set_fill(gtk.gdk.SOLID)
        
    # Fill the background        
    self.image.draw_rectangle(self.colorgc,True,0,0,self.rect.width,self.rect.height)
    
    if not self.sorted_sprites: return
    i = self.viewport_index
    for i in self.visible_sprite_indexes:
      #self.sorted_sprites[i].draw(self.image)    
      self.sorted_sprites[i].draw(self)   
    
    self.dirty = False
    self.window.end_paint() 
      
  def upbutton(self, widget, event, data=None):
      ''' User let go of the screen '''
      mousepoint = Rect((event.x,event.y))
      self.mousex = event.x
      self.mousey = event.y
      
      if self.mousedown:
        self.mousedown  = 0
        self.friction_c = 0
        if self._is_tap():
          # This was a tap, not a drag
          self._change_state('tap')
          self.stop_animation()
          
      if self.last_state == 'idle' and self.cur_state == 'tap': self.tap_select()
      self.dirty = True
      if len(self.sprites) == 0: return
      self.update()
      self.expose()   
      
  def clickbutton(self, widget, event, data=None):
      ''' User clicking down on the mouse '''
      mousepoint = (event.x,event.y)
      self.smousey = event.y
      
      if self.rect.collidepoint(mousepoint):
        self.tap_select(False)
        self.speed      = 0
        self.distance   = 0
        self.num_events = 0
        self.mousedown = 1
      self.dirty = True
      self.update()
      self.expose()   
      
    
  def mousemove(self, widget, event, data=None):
      ''' User dragging the widget around '''
      if event.is_hint:
          self.mousex, self.mousey, state = event.window.get_pointer()
      else:
          self.mousex = event.x
          self.mousey = event.y
          state = event.state
          
      if self.mousedown and self.sorted_sprites and self.is_scrollable():
        rel_y = int(self.mousey) - int(self.smousey)
        self.offset = self.offset + rel_y
        if cmp(rel_y,0) and cmp(rel_y,0) != cmp(self.distance,0): # changed direction
          self.num_events = 0
          self.distance   = 0
        self.num_events = self.num_events + 1
        self.distance   = self.distance   + rel_y
        #self.distance = rel_y
        self.speed      = self.sorted_sprites[0].rect.height / 50 * self.distance / self.num_events
                
        if not self._is_tap(): self._change_state('drag')        
        self.smousey = event.y
        self.dirty = True
      
  def _change_state(self,new_state):
    l = ['idle','tap','drag']
    if new_state not in l:
      print "WARNING: unknown state %s" % (new_state,)
      return
    self.last_state = self.cur_state
    self.cur_state  = new_state

  def _sort_sprites(self):
    if (self.need_sort): # cache sort order between list updates
      self.scrollable = None
      #print "sorting"
      self.need_sort = False
      self.sorted_sprites = self.sprites
      #self.sorted_sprites.sort()

  def add(self,*sprite):
    self.need_sort = True
    self.sprites.append(*sprite)

  def remove(self,*sprite):
    self.need_sort = True
    self.sprites.remove(*sprite)

  def has_selected(self):
    ''' The "selected" sprite is the one that we think the user has chosen '''
    if self.selected_index == None: return False
    return True

  def next(self):
    ''' Move selected_sprite to next sprite in list '''
    if self.is_idle():
      if self.has_selected():
        self.set_selected_index(self.selected_index + 1,True)
      else:
        self.new_selected_index()

  def prev(self):
    ''' Move selected_sprite to previous sprite in list '''
    if self.is_idle():
      if self.has_selected():
        self.set_selected_index(self.selected_index - 1,True)
      else:
        self.new_selected_index()

  def get_selected_sprite(self):
    if not self.has_selected(): return None
    return self.sorted_sprites[self.selected_index]

  def get_selected_sprite_index(self):
    if not self.has_selected(): return None
    return self.selected_index

  def _calc_new_index(self,index,list):
    length = len(list)
    while index != 0: # FIXME: looping is silly for very large abs() values of index. maybe use divmod instead?
      if index < 0: 
        index = index + length
      elif index > length - 1:
        index = index - length
      else: break # within the normal range, so return it
    return index

  def new_selected_index(self):
    self.selected_index = self.viewport_index
    self.offset = 0
    self.update()

  def set_selected_index(self,index,snap=False):
    if self.has_selected(): self.get_selected_sprite().set_focus(False)
    self.selected_index = self._calc_new_index(index,self.sorted_sprites)
    self.get_selected_sprite().set_focus(True)
    if snap and self.is_scrollable():
      if not self.is_index_mostly_visible(self.selected_index):
        self.viewport_index = self.selected_index
        self.offset = 0
    self.update()

  def set_selected_sprite(self,new_selected_sprite):
    c = 0
    for sprite in self.sorted_sprites: # make sure this new selected sprite exists in our list
      if new_selected_sprite == sprite:
        self.set_selected_index(c,True)
        return True
      c += 1
    print "invalid set_selected!" # this should never happen
    return False

  def get_sprites(self):
    self._sort_sprites()
    return self.sorted_sprites

  def _calc_viewport_index(self):
    while self.offset != 0:
      height = self.sorted_sprites[self.viewport_index].rect.height + self.y_spacer
      if self.offset > 0:
        if self.offset > height:
          self.offset = self.offset - height
          self.viewport_index = self._calc_new_index(self.viewport_index - 1,self.sorted_sprites)
        elif self.sorted_sprites[self.viewport_index].rect.top > self.rect.top:
          self.offset = 0 - height + self.offset
          self.viewport_index = self._calc_new_index(self.viewport_index - 1,self.sorted_sprites)
        else: return
      else:
        if abs(self.offset) > height:
          self.offset = self.offset + height
          self.viewport_index = self._calc_new_index(self.viewport_index + 1,self.sorted_sprites)
        else: return

  def _apply_friction(self):      
    self.friction_c = self.friction_c + 1
    if self.friction_c >= self.friction:
      self.friction_c = 0
      self.speed = self.speed - cmp(self.speed,0)
      
    if abs(self.speed) < 3: # Special handling to snap to a sprite when we get slow
        if abs(self.offset) > 5:
            # still have a ways to go, so don't slow down yet
            self.speed = self.speed + cmp(self.speed,0)
        else:
            # Just snap to the edge
            self.speed = 0
            self.offset = 0
              
    
  def _move(self):
    if not self.wrap: # prevent movement wrapping
      if ((self.speed < 0 or self.offset < 0) and self.viewport_index == len(self.sorted_sprites) - 1) or \
         ((self.speed > 0 or self.offset > 0) and self.viewport_index == 0):
        self.speed    = 0
        self.offset   = 0

    self.offset = self.offset + self.speed
    self._calc_viewport_index()
    if not self.speed: return
    self._apply_friction()

  def _update_sprite_rect(self,index,offset):      
    sprite = self.sorted_sprites[index]
    sprite.rect.top  = self.rect.top + offset
    sprite.rect.left = self.rect.left    
    
    if not self.rect.intersects(sprite.rect): return False
    return True

  def is_index_mostly_visible(self,index):
    if not self.sorted_sprites: return False
    if index not in self.visible_sprite_indexes: return False
    if self.rect.collidepoint(self.sorted_sprites[index].rect.center): return True
    return False

  def update(self):    
    # self.props.visible() test to see if I'm really here    
    
    if self.rect == None:
        self._makerect()
    self._move()
    self._sort_sprites()
    self.visible_sprite_indexes = []
    if not self.sorted_sprites: return

    i      = self.viewport_index
    offset = self.offset
    first  = True
    while self._update_sprite_rect(i,offset) or first:
      self.visible_sprite_indexes.append(i)
      offset = self.sorted_sprites[i].rect.bottom + self.y_spacer - self.rect.top
      if not self.wrap and i == len(self.sorted_sprites) - 1: break # prevent display wrapping
      i = self._calc_new_index(i+1,self.sorted_sprites)
      first = False
      if i == self.viewport_index: break # handle lists that are shorter than the viewport

    if self.has_selected():
      self.get_selected_sprite().update()

  def _is_tap(self):
    if self.num_events <= \
        self.tap_event_threshold and abs(self.distance) <= \
        self.tap_distance_threshold: return True
    return False

  def tap_select(self,run_callback=True):
    if not self.sorted_sprites: return
    mouse_pos = (self.mousex, self.mousey)
    for i in self.visible_sprite_indexes:
      if self.sorted_sprites[i].rect.collidepoint(mouse_pos):
        self.set_selected_index(i)
        if run_callback:
          if self.selected_cb <> None:
              self.selected_cb(self,None)
          self._change_state('idle')
        break

  def is_scrollable(self):
    if not self.visible_sprite_indexes: self.update()
    
    if self.scrollable != None: return self.scrollable
    if not self.sorted_sprites: return False
    height = 0
    for i in self.sorted_sprites:
      height += i.rect.height
    spacer_height = (len(self.visible_sprite_indexes) - 1) * self.y_spacer
    height += spacer_height
    
    if height > self.rect.height: self.scrollable = True
    else: self.scrollable = False
    return self.scrollable

  def is_idle(self):
    idle = (not self.speed and not self.mousedown)
    if idle and self.cur_state != 'idle': self._change_state('idle')
    return idle


  def stop_animation(self):
    self.speed      = 0
    self.distance   = 0
    self.num_events = 0
    self.mousedown  = 0
    self.friction_c = 0
    self.update()

  def close(self):
    for widget in self.sorted_sprites: widget.remove()
    